local ADDONNAME, A = select(1, ...);
local P = A.Player;
local C = A.Constants;
local U = A.Utils;
local PDB = A.PlayersDB;
local MSG = A.Message;
local TypeApi = A.Api.Type;
local Cache = A.Cache;

local Bus = CreateFrame("Frame");
A.ChatBus = Bus;

Bus.LFG_Prefix = A.CommunicationId;

local storeForInSeconds, cleanEachInSeconds = 10, 8;
local ChatThrottler = U:CreateThrottler(function(message) return message.time end, storeForInSeconds, cleanEachInSeconds);



local function normalizePlayerName(playerName)

    local index = string.find(playerName, "-");

    if index and not Bus.playerPostfix then
        local pstfx = U:TextSplit(playerName, "-");
        Bus.playerPostfix = U:JoinToString(pstfx, "-", 2);
        return playerName;
    elseif index then
        return playerName;
    elseif not Bus.playerPostfix then
        -- very rare case
        return;
    end

    -- here we go if there is no index and postfix detected

    return playerName .. "-" .. Bus.playerPostfix;

end

local function normalizeGuid(guid)
    if type(guid) ~= "string" or guid == "" then
        return;
    end
    
    local prefix = string.lower(strsub(guid, 1, 7));
    if prefix == "player-" then
        return guid;
    else
        --print(" # ?? ", guid, prefix);
    end
end



function Bus:GetLfgChannelId()
    local id = U:GetLFGChannelName();
    return id;
end

function Bus:SendChatMessage(text)
    local lfgChannelId = Bus:GetLfgChannelId();
    if not lfgChannelId then
        print("LookingForGroup channel not found. It seems you are not joined");
        return;
    end

    local howMany = floor(self.MessagesPerMinute / 75) + 1;
    for _i=1, howMany do
        SendChatMessage(text, "CHANNEL", nil, lfgChannelId);
    end
    --SendChatMessage(text, "CHANNEL", nil, lfgChannelId);
    print(U:InlineColor("was sent: ", "00aaff")..U:InlineColor(text, "ffdd00"));
    --SendChatMessage(text, "SAY");
end

function Bus:SendAddonMessage(text, player)
    local ret = C_ChatInfo.SendAddonMessage(Bus.LFG_Prefix, text, "WHISPER", player.name);
    if not ret then
        print('send internal message to ' .. player.name .. " fails");
    end
end

Bus.TypeHandlers = {};
Bus.AllHandlers = {};

Bus.MinuteStarts = GetTime();
Bus.TotalMessages = 0;
Bus.MinuteMessages = 0;
Bus.MessagesPerMinute = 0;

function Bus:Start()
    C_ChatInfo.RegisterAddonMessagePrefix(self.LFG_Prefix);
    self.started = true;
end

function Bus:GetTypeHolder(typeId)
    -- local holder = self.TypeHandlers[typeId];
    -- if not holder then
    --     holder = {};
    --     self.TypeHandlers[typeId] = holder;
    -- end
    -- return holder;
    return self.AllHandlers;
end

function Bus:AddTypeHandler(typeId, cb, sourceType, name)
    local holder = self:GetTypeHolder(typeId or "all");

    local handler = {
        name = name,
        typeId = typeId,
        cb = cb,
        sourceType,
    }

    table.insert(holder, handler);
    --print(" > added " .. name or "");
    return handler;
end

-- function Bus:RemoveTypeHandler(typeId, cb, sourceType)
--     local holder = self:GetTypeHolder(typeId);
--     table.remove(holder, handler);
-- end

function Bus:InvokeTypeCallbacksForMessage(sourceType, messageType, message)
    local typeId = messageType and messageType.id or "all";
    local holder = self:GetTypeHolder(typeId);
    for _, handler in pairs(holder) do

        if 
            (handler.typeId == nil or handler.typeId == typeId) 
            and 
            (handler.sourceType or sourceType) == sourceType 
        then
            --print("CALLBACK MATCHED",handler.name, " / ", handler.typeId, " / ", handler.sourceType, " / ");
            handler.cb(message);
        else
            --print(' - missed cb:', sourceType, " / ", messageType, " / ", handler.typeId, " / ", handler.sourceType, " / ");
        end
    end
end



-- function Bus:OnPlayerEnteringWorld(isLogin, isReload)
--     Cache:Initialize();
--     C_ChatInfo.RegisterAddonMessagePrefix(self.LFG_Prefix);
--     self.started = true;
--     A:Start();
-- end

-- should be filled
Bus.playerPostfix = nil;


local function invokeHandlers(handlers, ...)
    for _, handler in pairs(handlers) do
        if type(handler) == "function" then
            handler(...);
        elseif handler.callback then
            handler.callback(handler.context, ...);
        end
    end
end

function Bus:OnChatMsgAddon(comid, text, channel, playerName, playerName2)
    if not self.started then return end;

    playerName = normalizePlayerName(playerName);
    if not playerName then return end
    
    local message = MSG:CreateAddon(comid, text, channel, playerName, playerName2);

    if not message then return end;
    self:InvokeTypeCallbacksForMessage("addon", message.type, message);
end

function Bus:IncrementCounter()
    self.TotalMessages = self.TotalMessages + 1;
    self.MinuteMessages = self.MinuteMessages + 1;
end


function Bus:OnChatMsgChannel(text, playerName, languageName, channelName, playerName2, specialFlags, zoneChannelID, channelIndex, channelBaseName, unused, lineID, guid, bnSenderID, isMobile, isSubtitle, hideSenderInLetterbox, supressRaidIcons)
    -- drop message if we are not started
    if not self.started then return end;

    -- all messages should be counted for stats
    self:IncrementCounter();

    -- system and my own messages should be dropped
    if playerName == nil or playerName == "" or P:IsMyGuid(guid) then return end
    
    playerName = normalizePlayerName(playerName);
    if not playerName then
        --print(" - normalize fail:", playerName, "/", playerName2, "/", self.playerPostfix, ";");
        return;
    end
    
    -- drop message if exactly this message was posted in near 10 seconds
    if not ChatThrottler:TryAdd(playerName, true) then 
        --print(' < - throttled');
        return;
    end

    guid = normalizeGuid(guid);

    -- populating player info, without player level because this information is not available
    local player = PDB:GetOrCreate(playerName, "chat", { id = guid, name = playerName2, fullName = playerName });

    
    -- -- populating message object
    local message = MSG:Create(text, player);
    if not message then 
        --print(" < - no message");
        return 
    end;

    -- trying to detect if it is a valid message
    message:TryDetect();

    -- if not message.dungeon then
    --    print("> ", message.type and message.type.id, message.text);
    -- end

    -- invoke apropriate handlers for message based on message.type.id
    self:InvokeTypeCallbacksForMessage("chat", message.type, message);

    --print("< - # ", playerName, "  +++  ", playerName2, " +++ ", guid);
end


function Bus:OnEvent(event, ...)
    if event == "CHAT_MSG_ADDON" then
        self:OnChatMsgAddon(...);
    elseif event == "CHAT_MSG_CHANNEL" or event == "CHAT_MSG_SAY" then
        self:OnChatMsgChannel(...);
    -- elseif event == "PLAYER_ENTERING_WORLD" then
    --     local isLogin, isReload = ...;
    --     self:OnPlayerEnteringWorld(isLogin, isReload)
    end
end

Bus:RegisterEvent("CHAT_MSG_ADDON");
Bus:RegisterEvent("CHAT_MSG_CHANNEL");
Bus:RegisterEvent("CHAT_MSG_SAY");
--Bus:RegisterEvent("PLAYER_ENTERING_WORLD");
--Bus:RegisterEvent("ADDON_LOADED");
Bus:SetScript("OnEvent", Bus.OnEvent);


Bus._statsTimer = C_Timer.NewTicker(60, function()  
    Bus.MinuteStarts = GetTime();
    Bus.MessagesPerMinute = Bus.MinuteMessages;
    Bus.MinuteMessages = 0;
end);